############################################################################
'''
                        Pi Pico Simple Timer
    
    Started as an investigation of creating a simple timer similar to the water
tap timers that can be bought in a hardware stores.
    Parts:
        Pi Pico
        PiicoDev CAP1203 Sensor
        PiicoDev SSD1306 OLED
        Relay Module
    
    When it came to power saving, the Dormant mode of the Pico is used. But
the built in Micro Python routines do not result in power saving. The current
level did not change. Adafruit Circuit Python was tried and resulted in lower
current consumption but not to the levels specified in the RP2040 datasheet.
The Pi Pico C SDK library was tried and the levels were achieved. So the Pico
could be put into a power saving mode. LightSleep = 1.3mA and Dormant = 0.8mA.

    As the code was already working in MicroPython and the CAP1203 & SSD1306
PiicoDev libraries worked perfectly, a Python solution was sort. Investigation
of github revealed some development in MicroPython for the Pico low power.
    https://github.com/tomjorquera/pico-micropython-lowpower-workaround

   The github example switched the Pico into Dormant mode but it would never
leave. The code set Dormant mode but did not set the clocks as recommended
in the datasheet causing the Pico to lock up when returning from Dormant mode.
(best guess of what was observed)

    Using the example from the Pico C SDK library, the github example was heavily
modified to set the clocks as per the datasheet. All of the register manipulation
was placed in one Python def() statement. This was found to work more effectively
than using calls to other def() statements. Using calls the code appeared to work
initially, but if left for some time, failed to restart when triggered.  

    Program waiting sensor = 19.5mA
    Timer started and LCD on = 25mA
    Relay activated = 75mA
    Total when timer running about 100mA

    Dormant mode activated (after 5 minutes idle)
    Pi Pico = 0.3mA
    Pi Pico & CAP1203 = 0.7mA
    Pi Pico & CAP1203 & SSD1306 = 2.2mA
    Set LCD power off Total = 1.8mA

    The program now works correctly as a timer and switches into Dormant mode if
it has been idle for more than 5 minutes. The CAP1203 interrupt brings the Pico out
of Dormant mode and a timing interval can be selected.

    Created 06 Jan 2022, Final 16 Jan 2022
  
'''
############################################################################
import utime
from machine import Pin
from PiicoDev_Unified import sleep_ms
from PiicoDev_CAP1203 import PiicoDev_CAP1203
from PiicoDev_SSD1306 import *

############################################################################
# definitions
############################################################################
CAP_INT_PIN = 21
RELAY_PIN = 22
LED_PIN = 25

MINUTE_COUNT = 60000    # milliseconds 1 minute
BLANK_COUNT = 120000    # milliseconds 2 minutes
POWER_COUNT = 5         # minutes 5 minutes

# Testing only, faster timing 
#MINUTE_COUNT = 2000
#BLANK_COUNT = 4000
#POWER_COUNT = 5

############################################################################
# Address & Register definitions
############################################################################
# GPIO Register start address, GPIO pin determines actual register
#    calculated by int(gpio / 8) * 4
IO_BANK0_BASE = 0x40014000
# Register offsets
IO_BANK0_INTR0 = 0x0f0
IO_BANK0_DORMANT_WAKE_INTE0 = 0x160
# Bit determines type of interrupt, 4 bits per register, 8 registers in each 32 bit word
#     Bit position for GPIO is calculated by event << 4 * (gpio % 8)
#         event is one of 4 options below
LEVEL_LOW =  0x00000001
LEVEL_HIGH = 0x00000002
EDGE_FALL =  0x00000004
EDGE_RISE =  0x00000008

# XOSC Register start address
XOSC_BASE = 0x40024000
XOSC_DORMANT = 0x08
XOSC_STATUS = 0x04
# Values to write to set dormant mode, wake value is set by default
XOSC_DORMANT_VALUE = 0x636f6d61
XOSC_WAKE_VALUE = 0x77616b65
# bit to check indicating XOSC is stable
XOSC_STATUS_STABLE_BIT = 0x80000000
# ROSC Registers start address
ROSC_BASE = 0x40060000

# Clock Registers start address
CLOCKS_BASE = 0x40008000
# Register offsets
CLK_REF_CTRL = 0x30
CLK_SYS_CTRL = 0x3C
CLK_PERI_CTRL = 0x48
CLK_USB_CTRL = 0x54
CLK_ADC_CTRL = 0x60
CLK_RTC_CTRL = 0x6C
PLL_SYS_BASE = 0x40028000
PLL_USB_BASE = 0x4002c000
PLL_PWR = 0x4

############################################################################
# GPIO Pins
############################################################################
CAP1203_INT = Pin(CAP_INT_PIN, Pin.IN, Pin.PULL_UP)
Relay = Pin(RELAY_PIN, Pin.OUT)
LED = Pin(LED_PIN, Pin.OUT)

############################################################################
LED.value(1)        # turn on LED at start to indicate program started

############################################################################
# Create instances
############################################################################
CAP1203 = PiicoDev_CAP1203(sensitivity = 6)
display = create_PiicoDev_SSD1306()
display.fill(0)
display.show()
display.setContrast(0x1F)

############################################################################
# Variables
############################################################################
CountDown = 0                  # count down timer, minutes
CountDownPrev = CountDown      # previous value of count down timer, check for when count changed
TimeMinute = utime.ticks_ms()  # time at start of minute, milliseconds
LCD_TimeOn = utime.ticks_ms()  # time display active, milliseconds, blank after 2 minutes
PWR_Down = 0                   # counter to put Pico in Dormant mode, 5 minutes
Active = False                 # indicates relay is on
OnAlways = False               # indicates no timing, relay on all the time
LED_state = False              # 
switch = [(1,0),(2,0),(3,0)]   # dictionary of CAP1203 switch conditions 

############################################################################
# Assember routines to directly access registers
############################################################################
@micropython.asm_thumb
def _read_bits(r0):
    ldr(r0, [r0, 0])

@micropython.asm_thumb
def _write_bits(r0, r1):
    str(r1, [r0, 0])

############################################################################
# Set dormant mode and activate
#     Found it better to put all the register manipulation in the same Python def()
############################################################################
def dormant_until_pin(gpio_pin, edge=True, level=True):
# convert edge and level to event
    if not edge and not level:
        event = LEVEL_LOW
    if not edge and level:
        event = LEVEL_HIGH
    if edge and not level:
        event = EDGE_FALL
    if edge and level:
        event = EDGE_RISE
#check for valid gpio
    if gpio_pin < 0 or gpio_pin > 27:
        raise RuntimeError("Invalid value for pin " + str(gpio_pin) + " (expect int between 0 and 27)")
# clear interrrupts
    _write_bits(IO_BANK0_BASE + IO_BANK0_INTR0 + int(gpio_pin / 8) * 4, event << 4 * (gpio_pin % 8))
# Enable Wake-up from GPIO IRQ
    _write_bits(IO_BANK0_BASE + IO_BANK0_DORMANT_WAKE_INTE0 + int(gpio_pin / 8) * 4, event << 4 * (gpio_pin % 8))
# Setup clocks for dormant
    _write_bits(CLOCKS_BASE + CLK_REF_CTRL, 0x02)
    _write_bits(CLOCKS_BASE + CLK_SYS_CTRL, 0x00)
    _write_bits(CLOCKS_BASE + CLK_USB_CTRL, 0x00)
    _write_bits(CLOCKS_BASE + CLK_ADC_CTRL, 0x00)
    _write_bits(CLOCKS_BASE + CLK_RTC_CTRL, 0x60)
    _write_bits(CLOCKS_BASE + CLK_PERI_CTRL, 0x00)
    _write_bits(PLL_USB_BASE + PLL_PWR, 0x2D)
    _write_bits(ROSC_BASE, 0xD1EAA0)
# go dormant
    _write_bits(XOSC_BASE + XOSC_DORMANT, XOSC_DORMANT_VALUE)
####################################
# XXXX execution stops here XXXX
####################################
# Normalise clocks
    _write_bits(CLOCKS_BASE + CLK_REF_CTRL, 0x02)
    _write_bits(CLOCKS_BASE + CLK_SYS_CTRL, 0x01)
    _write_bits(CLOCKS_BASE + CLK_USB_CTRL, 0x800)
    _write_bits(CLOCKS_BASE + CLK_ADC_CTRL, 0x800)
    _write_bits(CLOCKS_BASE + CLK_RTC_CTRL, 0x800)
    _write_bits(CLOCKS_BASE + CLK_PERI_CTRL, 0x800)
    _write_bits(PLL_USB_BASE + PLL_PWR, 0x04)
    _write_bits(ROSC_BASE, 0xfffaa0)
# turn LED on to indicate out or dormant
    LED.value(1)
# Check for stable XOSC after wake up
    while not _read_bits(XOSC_BASE + XOSC_STATUS) & XOSC_STATUS_STABLE_BIT:
         pass
# clear interrupts
    _write_bits(IO_BANK0_BASE + IO_BANK0_INTR0 + int(gpio_pin / 8) * 4, event << 4 * (gpio_pin % 8))

############################################################################
# Flash LED
############################################################################
def LED_Flash(n):
    for _ in range(n):
        LED.value(1)
        sleep_ms(100)
        LED.value(0)
        sleep_ms(100)

############################################################################
# LCD bitmaps to display
############################################################################
def LCD_Display():
# note: display blanks in between intervals
    display.fill(0)
    if OnAlways:
        display.load_pbm('bitmaps/ON.pbm', 1)
    else:
        if CountDown == 5:
            display.load_pbm('bitmaps/5.pbm', 1)
        elif CountDown == 10:
            display.load_pbm('bitmaps/10.pbm', 1)
        elif CountDown == 15:
            display.load_pbm('bitmaps/15.pbm', 1)
        elif CountDown == 20:
            display.load_pbm('bitmaps/20.pbm', 1)
        elif CountDown == 25:
            display.load_pbm('bitmaps/25.pbm', 1)
        elif CountDown == 30:
            display.load_pbm('bitmaps/30.pbm', 1)
        elif CountDown == 35:
            display.load_pbm('bitmaps/30.pbm', 1)
        elif CountDown == 40:
            display.load_pbm('bitmaps/40.pbm', 1)
        elif CountDown == 45:
            display.load_pbm('bitmaps/45.pbm', 1)
        elif CountDown == 50:
            display.load_pbm('bitmaps/50.pbm', 1)
        elif CountDown == 55:
            display.load_pbm('bitmaps/55.pbm', 1)
        elif CountDown == 60:
            display.load_pbm('bitmaps/60.pbm', 1)
        elif CountDown == 65:
            display.load_pbm('bitmaps/65.pbm', 1)
        elif CountDown == 70:
            display.load_pbm('bitmaps/70.pbm', 1)
        elif CountDown == 75:
            display.load_pbm('bitmaps/75.pbm', 1)
        elif CountDown == 80:
            display.load_pbm('bitmaps/80.pbm', 1)
        elif CountDown == 85:
            display.load_pbm('bitmaps/85.pbm', 1)
        elif CountDown == 90:
            display.load_pbm('bitmaps/90.pbm', 1)
        elif CountDown == 95:
            display.load_pbm('bitmaps/95.pbm', 1)
        elif CountDown == 100:
            display.load_pbm('bitmaps/100.pbm', 1)
        elif CountDown == 105:
            display.load_pbm('bitmaps/105.pbm', 1)
        elif CountDown == 110:
            display.load_pbm('bitmaps/110.pbm', 1)
        elif CountDown == 115:
            display.load_pbm('bitmaps/115.pbm', 1)
        elif CountDown == 120:
            display.load_pbm('bitmaps/120.pbm', 1)
    display.show()
    return

############################################################################
# Main Loop
############################################################################
while True:
    LED.value(0)         # Turn LED off when in main loop
############################################################################
# Check CAP1203 semsor, if interrupt, read CAP1203 and check switch press
# Switch 1 cycles through time intervals, ON condition when at 0
# Switch 2 activates, if deactivated timer set to 0 to turn off relay
############################################################################
    if CAP1203_INT.value() == 0:
        CAP1203.clearInterrupt()
        switch = CAP1203.read()
        if switch[1] == 1:
            if not Active:
                if CountDown < 120:
                    CountDown += 30
                    TimeMinute = utime.ticks_ms()
                    OnAlways = False
                else:
                    CountDown = 0
                    OnAlways = True
                LCD_Display()
            PWR_Down = 0
        if switch[2] == 1:
            if OnAlways | CountDown > 0:
                if Active:
                    CountDown = 0
                    PWR_Down = 0
                    OnAlways = False
                    LED.value(0)
                    LED_state = False
                    LCD_Display()
                    Active = False
                else:
                    Active = True

############################################################################
# Count millisecs till 1 minute
#     if active  update CountDown
#     if not increment power down counter
############################################################################
    if not OnAlways:
        if utime.ticks_diff(utime.ticks_ms(), TimeMinute) > MINUTE_COUNT:
            TimeMinute = utime.ticks_ms()
            if Active:
                if CountDown > 0:
                    CountDown -= 1
                else:
                    Active = False
            else:
                PWR_Down += 1

############################################################################# if active Flash LED to indicate timer is running
############################################################################
    if Active:
        if LED_state:
            LED.value(0)
            LED_state = False
        else:
            LED.value(1)
            LED_state = True

############################################################################
# if count down changed update display
############################################################################
    if CountDownPrev != CountDown:
        LCD_Display()
        LCD_TimeOn = utime.ticks_ms()
        CountDownPrev = CountDown

############################################################################
# Blank display after 2 minutes, if active
# If not active reset timer for when switch pressed again
#     So switch press with blank screen will always starts at 30
############################################################################
    if utime.ticks_diff(utime.ticks_ms(), LCD_TimeOn) > BLANK_COUNT:
        if not Active:
            CountDown = 0
        display.fill(0)
        display.show()

############################################################################
# Power down. Puts Pico into Dormant mode if no activity for greater than 5 minutes
############################################################################
    if not OnAlways:
        if not Active:
            if PWR_Down > POWER_COUNT:
                PWR_Down = 0
                CAP1203.clearInterrupt()
                display.fill(0)
                display.show()
                display.poweroff()
                LED_Flash(4)        # Flash LED to indicate Dormant mode activated.
                LED.value(0)
                dormant_until_pin(CAP_INT_PIN, True, False)   # edge Level
# XXXX Program stops here XXXX
                LED.value(1)        # turn LED on to indicate out of Dormant mode
                sleep_ms(200)
                sleep_ms(2000)
                display.poweron()

############################################################################
# Relay Control
#     When ON condition, switch 2 turns relay on/off
#     When timer condition, switch 2 activates and can turn off, resetting condition
#     When timer reaches 0, relay turned off
############################################################################
    if OnAways:
        if Active:
            Relay.on()
        else:
            Relay.off()
    else:      
        if CountDown > 0:
            if Active:
                Relay.on()
            else:
                Relay.off()
        else:
            Relay.off()
        
    sleep_ms(100)
